@@ -35,8 +35,12 @@
 #include <glib.h>
 #include <libgnome/libgnome.h>
 #include <libgnomevfs/gnome-vfs.h>
+#include <libgnomevfs/gnome-vfs-handle.h>
 #include <libgnomevfs/gnome-vfs-mime.h>
+#include <libgnomevfs/gnome-vfs-result.h>
+#include <gconf/gconf-client.h>
 #include "gthumb-init.h"
+#include "gconf-utils.h"
 #include "file-utils.h"
 
 
@@ -56,7 +60,10 @@ path_list_data_new ()
 	pli->files = NULL;
 	pli->dirs = NULL;
 	pli->done_func = NULL;
-	pli->data = NULL;
+	pli->done_data = NULL;
+	pli->interrupt_func = NULL;
+	pli->interrupt_data = NULL;
+	pli->interrupted = FALSE;
 
 	return pli;
 }
@@ -101,32 +108,39 @@ directory_load_cb (GnomeVFSAsyncHandle *handle,
 		   gpointer data)
 {
 	PathListData *pli;
-	GnomeVFSFileInfo *info;
 	GList *node;
 
 	pli = (PathListData *) data;
 	pli->result = result;
 
-	for (node = list; node != NULL; node = node->next) {
-		GnomeVFSURI *full_uri = NULL;
-		gchar *str_uri, *unesc_uri;
+	if (pli->interrupted) {
+		if (pli->interrupt_func) 
+			pli->interrupt_func (pli->interrupt_data);
+		path_list_data_free (pli);
+		return;
+	}
 
-		info = node->data;
+	for (node = list; node != NULL; node = node->next) {
+		GnomeVFSFileInfo * info     = node->data;
+		GnomeVFSURI *      full_uri = NULL;
+		gchar *            str_uri;
+		gchar *            unesc_uri;
 
 		switch (info->type) {
 		case GNOME_VFS_FILE_TYPE_REGULAR:
-			full_uri = gnome_vfs_uri_append_file_name (pli->uri, 
-								   info->name);
+			full_uri = gnome_vfs_uri_append_file_name (pli->uri, info->name);
 			str_uri = gnome_vfs_uri_to_string (full_uri, GNOME_VFS_URI_HIDE_TOPLEVEL_METHOD);
 			unesc_uri = gnome_vfs_unescape_string (str_uri, NULL);
-			
+
 			pli->files = g_list_prepend (pli->files, unesc_uri);
 			g_free (str_uri);
 			break;
 
 		case GNOME_VFS_FILE_TYPE_DIRECTORY:
-			full_uri = gnome_vfs_uri_append_path (pli->uri, 
-							      info->name);
+			if (SPECIAL_DIR (info->name))
+				break;
+
+			full_uri = gnome_vfs_uri_append_path (pli->uri, info->name);
 			str_uri = gnome_vfs_uri_to_string (full_uri, GNOME_VFS_URI_HIDE_TOPLEVEL_METHOD);
 			unesc_uri = gnome_vfs_unescape_string (str_uri, NULL);
 
@@ -135,6 +149,7 @@ directory_load_cb (GnomeVFSAsyncHandle *handle,
 			break;
 
 		default:
+			break;
 		}
 
 		if (full_uri)
@@ -145,7 +160,7 @@ directory_load_cb (GnomeVFSAsyncHandle *handle,
 	    || (result != GNOME_VFS_OK)) {
 		if (pli->done_func) 
 			/* pli is deallocated in pli->done_func */
-			pli->done_func (pli, pli->data);
+			pli->done_func (pli, pli->done_data);
 		else
 			path_list_data_free (pli);
 
@@ -155,33 +170,30 @@ directory_load_cb (GnomeVFSAsyncHandle *handle,
 
 
 PathListHandle *
-path_list_async_new (const gchar *uri, 
-		     PathListDoneFunc f,
-		     gpointer data)
+path_list_async_new (const gchar      *uri, 
+		     PathListDoneFunc  f,
+		     gpointer         data)
 {
 	GnomeVFSAsyncHandle *handle;
 	PathListData *pli;
 	gchar *escaped;
 	PathListHandle *pl_handle;
 
-	escaped = gnome_vfs_escape_path_string (uri);
-
 	pli = path_list_data_new ();
-	pli->uri = gnome_vfs_uri_new (escaped);
-	pli->done_func = f;
-	pli->data = data;
 
+	escaped = gnome_vfs_escape_path_string (uri);
+	pli->uri = gnome_vfs_uri_new (escaped);
 	g_free (escaped);
 
+	pli->done_func = f;
+	pli->done_data = data;
+
 	gnome_vfs_async_load_directory_uri (
 		&handle,
 		pli->uri,
 		GNOME_VFS_FILE_INFO_FOLLOW_LINKS,
-		GNOME_VFS_DIRECTORY_FILTER_NONE,
-		(GNOME_VFS_DIRECTORY_FILTER_NOSELFDIR 
-		 | GNOME_VFS_DIRECTORY_FILTER_NOPARENTDIR), 
-		NULL,
 		128 /* items_per_notification FIXME */,
+		GNOME_VFS_PRIORITY_DEFAULT,
 		directory_load_cb,
 		pli);
 
@@ -194,10 +206,14 @@ path_list_async_new (const gchar *uri,
 
 
 void
-path_list_async_interrupt (PathListHandle *handle)
+path_list_async_interrupt (PathListHandle   *handle,
+			   DoneFunc          f,
+			   gpointer          data)
 {
-	gnome_vfs_async_cancel (handle->vfs_handle);
-	path_list_data_free (handle->pli_data);
+	handle->pli_data->interrupted = TRUE;
+	handle->pli_data->interrupt_func = f;
+	handle->pli_data->interrupt_data = data;
+
 	g_free (handle);
 }
 
@@ -445,8 +461,49 @@ visit_rc_directory_async (const gchar *rc_dir,
 }
 
 
+/* -- rmdir_recursive -- */
+
+
+gboolean
+rmdir_recursive (const gchar *directory)
+{
+	GList    *files, *dirs;
+	GList    *scan;
+	gboolean  error = FALSE;
+
+	if (! path_is_dir (directory)) 
+		return FALSE;
+
+	path_list_new (directory, &files, &dirs);
+
+	for (scan = files; scan; scan = scan->next) {
+		char *file = scan->data;
+		if ((unlink (file) < 0)) {
+			g_warning ("Cannot delete %s\n", file);
+			error = TRUE;
+		}
+	}
+	path_list_free (files);
+
+	for (scan = dirs; scan; scan = scan->next) {
+		char *sub_dir = scan->data;
+		if (rmdir_recursive (sub_dir) == FALSE)
+			error = TRUE;
+		if (rmdir (sub_dir) == 0)
+			error = TRUE;
+	}
+	path_list_free (dirs);
+
+	if (rmdir (directory) == 0)
+		error = TRUE;
+
+	return !error;
+}
+
+
 /* -- */
 
+
 gboolean
 path_is_file (const gchar *path)
 {
@@ -475,14 +532,15 @@ path_is_file (const gchar *path)
 
 
 gboolean
-path_is_dir (const gchar *path)
+path_is_dir (const char *path)
 {
 	GnomeVFSFileInfo *info;
-	GnomeVFSResult result;
-	gboolean is_dir;
-	gchar *escaped;
+	GnomeVFSResult    result;
+	gboolean          is_dir;
+	char             *escaped;
 
-	if (! path || ! *path) return FALSE; 
+	if (! path || ! *path) 
+		return FALSE; 
 
 	info = gnome_vfs_file_info_new ();
 	escaped = gnome_vfs_escape_path_string (path);
@@ -526,12 +584,12 @@ dir_is_empty (const gchar *path)
 
 
 GnomeVFSFileSize 
-get_file_size (const gchar *path)
+get_file_size (const char *path)
 {
 	GnomeVFSFileInfo *info;
-	GnomeVFSResult result;
-	GnomeVFSFileSize size;
-	gchar *escaped;
+	GnomeVFSResult    result;
+	GnomeVFSFileSize  size;
+	char             *escaped;
 
 	if (! path || ! *path) return 0; 
 
@@ -553,12 +611,12 @@ get_file_size (const gchar *path)
 
 
 time_t 
-get_file_mtime (const gchar *path)
+get_file_mtime (const char *path)
 {
 	GnomeVFSFileInfo *info;
-	GnomeVFSResult result;
-	gchar *escaped;
-	time_t mtime;
+	GnomeVFSResult    result;
+	char             *escaped;
+	time_t            mtime;
 
 	if (! path || ! *path) return 0; 
 
@@ -569,7 +627,8 @@ get_file_mtime (const gchar *path)
 					  (GNOME_VFS_FILE_INFO_DEFAULT 
 					   | GNOME_VFS_FILE_INFO_FOLLOW_LINKS));
 	mtime = 0;
-	if (result == GNOME_VFS_OK)
+	if ((result == GNOME_VFS_OK)
+	    && (info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_MTIME))
 		mtime = info->mtime;
 
 	g_free (escaped);
@@ -596,7 +655,8 @@ get_file_ctime (const gchar *path)
 					  (GNOME_VFS_FILE_INFO_DEFAULT 
 					   | GNOME_VFS_FILE_INFO_FOLLOW_LINKS));
 	ctime = 0;
-	if (result == GNOME_VFS_OK)
+	if ((result == GNOME_VFS_OK)
+	    && (info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_CTIME))
 		ctime = info->ctime;
 
 	g_free (escaped);
@@ -606,6 +666,26 @@ get_file_ctime (const gchar *path)
 }
 
 
+void
+set_file_mtime (const gchar *path,
+		time_t       mtime)
+{
+	GnomeVFSFileInfo *file_info;
+	char             *escaped_path;
+
+	file_info = gnome_vfs_file_info_new ();
+	file_info->mtime = mtime;
+	file_info->atime = mtime;
+
+	escaped_path = gnome_vfs_escape_path_string (path);
+	gnome_vfs_set_file_info (escaped_path,
+				 file_info,
+				 GNOME_VFS_SET_FILE_INFO_TIME);
+	gnome_vfs_file_info_unref (file_info);
+	g_free (escaped_path);
+}
+
+
 gboolean 
 file_copy (const gchar *from, 
 	   const gchar *to)
@@ -665,13 +745,10 @@ file_move (const gchar *from,
 }
 
 
-#define SPECIAL_DIR(x) (! strcmp (x, "..") || ! strcmp (x, "."))
-
-
 gboolean 
-path_list_new (const gchar *path, 
-	       GList **files, 
-	       GList **dirs)
+path_list_new (const char  *path, 
+	       GList      **files, 
+	       GList      **dirs)
 {
 	DIR *dp;
 	struct dirent *dir;
@@ -719,7 +796,19 @@ path_list_new (const gchar *path,
 	return TRUE;
 }
 
-#undef SPECIAL_DIR
+
+GList *
+path_list_dup (GList *path_list)
+{
+	GList *new_list = NULL;
+	GList *scan;
+
+	for (scan = path_list; scan; scan = scan->next)
+		new_list = g_list_prepend (new_list, g_strdup (scan->data));
+
+	return g_list_reverse (new_list);
+}
+
 
 void 
 path_list_free (GList *list)
@@ -731,6 +820,18 @@ path_list_free (GList *list)
 }
 
 
+GList *
+path_list_find_path (GList *list, const char *path)
+{
+	GList *scan;
+
+	for (scan = list; scan; scan = scan->next)
+		if (strcmp ((char*) scan->data, path) == 0)
+			return scan;
+	return NULL;
+}
+
+
 gboolean 
 file_is_hidden (const gchar *name)
 {
@@ -744,15 +845,15 @@ file_is_hidden (const gchar *name)
 
 gboolean
 file_is_image (const gchar *name,
-	       gboolean fast_file_type)
+	       gboolean     fast_file_type)
 {
 	const char *result;
-	gboolean is_an_image;
+	gboolean    is_an_image;
 
 	if (fast_file_type)
 		result = gnome_vfs_mime_type_from_name_or_default (name, NULL);
 	else 
-		result = gnome_mime_type_from_magic (name);
+		result = gnome_vfs_get_file_mime_type (name, NULL, FALSE);
 
 	/* Unknown file type. */
 	if (result == NULL)
@@ -766,6 +867,32 @@ file_is_image (const gchar *name,
 }
 
 
+gboolean 
+image_is_jpeg (const char *name)
+{
+	const char *result;
+
+	if (eel_gconf_get_boolean (PREF_FAST_FILE_TYPE))
+		result = gnome_vfs_mime_type_from_name_or_default (name, NULL);
+	else 
+		result = gnome_vfs_get_file_mime_type (name, NULL, FALSE);
+	
+	/* Unknown file type. */
+	if (result == NULL)
+		return FALSE;
+
+	return (strcmp (result, "image/jpeg") == 0);
+}
+
+
+gboolean
+file_extension_is (const char *filename, 
+		   const char *ext)
+{
+	return ! strcasecmp (filename + strlen (filename) - strlen (ext), ext);
+}
+
+
 long 
 checksum_simple (const gchar *path)
 {
@@ -784,12 +911,30 @@ checksum_simple (const gchar *path)
 }
 
 
-const gchar *
-file_name_from_path (const gchar *path)
+/* like g_basename but does not warns about NULL and does not
+ * alloc a new string. */
+G_CONST_RETURN gchar *
+file_name_from_path (const gchar *file_name)
 {
-	/* g_basename warns about NULL, but we accept it. */
-	if (! path) return NULL;
-	return g_basename (path);
+	register gssize base;
+        register gssize last_char;
+
+        if (file_name == NULL)
+                return NULL;
+
+        if (file_name[0] == '\0')
+                return "";
+
+        last_char = strlen (file_name) - 1;
+
+        if (file_name [last_char] == G_DIR_SEPARATOR)
+                return "";
+
+        base = last_char;
+        while ((base >= 0) && (file_name [base] != G_DIR_SEPARATOR))
+                base--;
+
+        return file_name + base + 1;
 }
 
 
@@ -800,13 +945,17 @@ remove_level_from_path (const gchar *path)
 	const gchar *ptr = path;
 	gint p;
 
-	if (! path) return NULL;
+	if (! path) 
+		return NULL;
 
 	p = strlen (path) - 1;
-	if (p < 0) return NULL;
+	if (p < 0) 
+		return NULL;
 
-	while ((ptr[p] != '/') && (p > 0)) p--;
-	if ((p == 0) && (ptr[p] == '/')) p++;
+	while ((p > 0) && (ptr[p] != '/')) 
+		p--;
+	if ((p == 0) && (ptr[p] == '/')) 
+		p++;
 	new_path = g_strndup (path, (guint)p);
 
 	return new_path;
@@ -817,15 +966,20 @@ gchar *
 remove_extension_from_path (const gchar *path)
 {
 	gchar *new_path;
+	gint len;
 	const gchar *ptr = path;
 	gint p;
 
-	if (!path) return NULL;
-	if (strlen (path) < 2) return g_strdup( path);
+	if (! path) 
+		return NULL;
+	len = strlen (path);
+	if (len == 1) 
+		return g_strdup (path);
 
-	p = strlen (path) - 1;
+	p = len - 1;
 	while ((ptr[p] != '.') && (p > 0)) p--;
-	if (p == 0) p = strlen (path) - 1;
+	if (p == 0) 
+		p = len;
 	new_path = g_strndup (path, (guint) p);
 
 	return new_path;
@@ -916,8 +1070,9 @@ dir_list_filter_and_sort (GList *dir_list,
 
 
 /* characters to escape */
-static gchar bad_char[] = { '$', '\'', '`', '"', '\\', '!', '?', 
-			    ' ', '(', ')', '[', ']', '&', '|' };
+static gchar bad_char[] = { '$', '\'', '`', '"', '\\', '!', '?', '*',
+			    ' ', '(', ')', '[', ']', '&', '|', '@' , '#',
+			    ';' };
 
 /* the size of bad_char */
 static const gint bad_chars = sizeof (bad_char) / sizeof (gchar);
@@ -973,3 +1128,250 @@ shell_escape (const gchar *filename)
 
 	return escaped;
 }
+
+
+char *
+escape_underscore (const char *name)
+{
+	const char *s;
+	char       *e_name, *t;
+	int         l = 0, us = 0;
+
+	if (name == NULL)
+		return NULL;
+
+	for (s = name; *s != 0; s++) {
+		if (*s == '_')
+			us++;
+		l++;
+	}
+        
+	if (us == 0)
+		return g_strdup (name);
+
+	e_name = g_malloc (sizeof (char) * (l + us + 1));
+
+	t = e_name;
+	for (s = name; *s != 0; s++) 
+		if (*s == '_') {
+			*t++ = '_';
+			*t++ = '_';
+		} else
+			*t++ = *s;
+	*t = 0;
+
+	return e_name;
+}
+
+
+char *
+get_terminal (gboolean with_exec_flag)
+{
+	GConfClient *client;
+	char        *result;
+	char        *terminal = NULL;
+	char        *exec_flag = NULL;
+
+	client = gconf_client_get_default ();
+	terminal = gconf_client_get_string (client, "/desktop/gnome/applications/terminal/exec", NULL);
+	g_object_unref (G_OBJECT (client));
+	
+	if (terminal) 
+		exec_flag = gconf_client_get_string (client, "/desktop/gnome/applications/terminal/exec_arg", NULL);
+
+	if (terminal == NULL) {
+		char *check;
+
+		check = g_find_program_in_path ("gnome-terminal");
+		if (check != NULL) {
+			terminal = check;
+			/* Note that gnome-terminal takes -x and
+			 * as -e in gnome-terminal is broken we use that. */
+			exec_flag = g_strdup ("-x");
+		} else {
+			if (check == NULL)
+				check = g_find_program_in_path ("nxterm");
+			if (check == NULL)
+				check = g_find_program_in_path ("color-xterm");
+			if (check == NULL)
+				check = g_find_program_in_path ("rxvt");
+			if (check == NULL)
+				check = g_find_program_in_path ("xterm");
+			if (check == NULL)
+				check = g_find_program_in_path ("dtterm");
+			if (check == NULL) {
+				g_warning ("Cannot find a terminal, using "
+					   "xterm, even if it may not work");
+				check = g_strdup ("xterm");
+			}
+			terminal = check;
+			exec_flag = g_strdup ("-e");
+		}
+	}
+
+	if (terminal == NULL)
+		return NULL;
+
+	if (with_exec_flag)
+		result = g_strconcat (terminal, " ", exec_flag, NULL);
+	else
+		result = g_strdup (terminal);
+
+	return result;
+}
+
+
+gchar *
+application_get_command (const GnomeVFSMimeApplication *app)
+{
+	char *command;
+
+	if (app->requires_terminal) {
+		char *terminal;
+		char *command_to_exec;
+
+		terminal = get_terminal (TRUE);
+		if (terminal == NULL)
+			return NULL;
+
+		command_to_exec = shell_escape (app->command);
+		command = g_strconcat (terminal,
+				       " ",
+				       command_to_exec,
+				       NULL);
+		g_free (terminal);
+		g_free (command_to_exec);
+	} else
+		command = shell_escape (app->command);
+
+	return command;
+}
+
+
+/* example 1 : filename = /xxx/yyy/zzz/foo 
+ *             destdir  = /xxx/www 
+ *             return   : ../yyy/zzz/foo
+ *
+ * example 2 : filename = /xxx/yyy/foo
+ *             destdir  = /xxx
+ *             return   : yyy/foo
+ */
+char *
+get_path_relative_to_dir (const char *filename, 
+			  const char *destdir)
+{
+	char     *sourcedir;
+	char    **sourcedir_v;
+	char    **destdir_v;
+	int       i, j;
+	char     *result;
+	GString  *relpath;
+
+	sourcedir = remove_level_from_path (filename);
+	sourcedir_v = g_strsplit (sourcedir, "/", 0);
+	destdir_v = g_strsplit (destdir, "/", 0);
+
+	relpath = g_string_new (NULL);
+
+	i = 0;
+	while ((sourcedir_v[i] != NULL)
+	       && (destdir_v[i] != NULL)
+	       && (strcmp (sourcedir_v[i], destdir_v[i]) == 0))
+		i++;
+	
+	j = i;
+
+	while (destdir_v[i++] != NULL) 
+		g_string_append (relpath, "../");
+
+	while (sourcedir_v[j] != NULL) {
+		g_string_append (relpath, sourcedir_v[j]);
+		g_string_append_c (relpath, '/');
+		j++;
+	}
+
+	g_string_append (relpath, file_name_from_path (filename));
+
+	g_strfreev (sourcedir_v);
+	g_strfreev (destdir_v);
+	g_free (sourcedir);
+
+	result = relpath->str;
+	g_string_free (relpath, FALSE);
+
+	return result;
+}
+
+
+/*
+ * example 1) input : "/xxx/yyy/.."         output : "/xxx"
+ * example 2) input : "/xxx/yyy/../www"     output : "/xxx/www"
+ * example 3) input : "/xxx/../yyy/../www"  output : "/www"
+ * example 4) input : "/xxx/../yyy/../"     output : "/"
+ * example 5) input : "/xxx/./"             output : "/xxx"
+ * example 6) input : "/xxx/./yyy"          output : "/xxx/yyy"
+ *
+ * Note : PATH must be absolute.
+ */
+char *
+remove_special_dirs_from_path (const char *path)
+{
+	char    **pathv;
+	GList    *list = NULL, *scan;
+	int       i;
+	GString  *result_s;
+	char     *result;
+
+	if ((path == NULL) || (*path != '/'))
+		return NULL;
+
+	if (strstr (path, ".") == NULL)
+		return g_strdup (path);
+
+	pathv = g_strsplit (path, "/", 0);
+
+	/* start from 1 to remove the first / that will be re-added later. */
+	for (i = 1; pathv[i] != NULL; i++) {
+		if (strcmp (pathv[i], ".") == 0) {
+			/* nothing to do. */
+		} else if (strcmp (pathv[i], "..") == 0) {
+			if (list == NULL) {
+				/* path error. */
+				g_strfreev (pathv);
+				return NULL;
+			}
+			list = g_list_delete_link (list, list);
+		} else 
+			list = g_list_prepend (list, pathv[i]);
+	}
+
+	result_s = g_string_new (NULL);
+	if (list == NULL)
+		g_string_append_c (result_s, '/');
+	else {
+		list = g_list_reverse (list);
+		for (scan = list; scan; scan = scan->next) {
+			g_string_append_c (result_s, '/');
+			g_string_append (result_s, scan->data);
+		}
+	}
+	result = result_s->str;
+	g_string_free (result_s, FALSE);
+	g_strfreev (pathv);
+
+	return result;
+}
+
+
+GnomeVFSURI *
+new_uri_from_path (const char *path)
+{
+	char        *escaped;
+	GnomeVFSURI *uri;
+
+	escaped = gnome_vfs_escape_path_string (path);
+	uri = gnome_vfs_uri_new (escaped);
+	g_free (escaped);
+
+	return uri;
+}
